// ================================================================
// LinBox Project 1999
// Symmetrizing iterator (for rank computations)
// Same left and right vector
// A is supposed to have tranpose-vector product
// the sequence is u^t u, (A u)^t (A u) = u^t (A^t A) u, 
// (A^t (A u))^t (A^t (A u)) = u^t (A^t A)^2 u , etc.
// Time-stamp: <16 Apr 03 13:25:54 Jean-Guillaume.Dumas@imag.fr> 
// ================================================================


#ifndef __BBContainer_SYMMETRIZE_H__
#define __BBContainer_SYMMETRIZE_H__
#include <LinBox/lin_rand.h>
#include <LinBox/lin_base_bbit.h>

template<class BlackBoxDomain, class Vecteur = typename BlackBoxDomain::PreferredInMatrix_t, class RandIter = Random>
class BB_Symmetrize_Container : public Base_BB_Container< BlackBoxDomain, Vecteur > {
    typedef Base_BB_Container< BlackBoxDomain, Vecteur > Father_t;
public:
    BB_Symmetrize_Container() {} 

    BB_Symmetrize_Container(BlackBoxDomain * D, const Vecteur& u0) 
            : Base_BB_Container< BlackBoxDomain, Vecteur>(D) { this->init(u0); }
    
    BB_Symmetrize_Container(BlackBoxDomain * D, RandIter& g ) 
            : Base_BB_Container< BlackBoxDomain, Vecteur>(D) { this->init(g); }
    
private:
    void _launch () {
        if (Father_t::even) {
            Father_t::even = 0;
            Father_t::_BB_domain->Apply(Father_t::v, Father_t::u);
            Father_t::DOTPROD(Father_t::_value,Father_t::v,Father_t::v); 
        } else {
            Father_t::even = 1;
            Father_t::_BB_domain->ApplyTrans( Father_t::u, Father_t::v); 
            Father_t::DOTPROD(Father_t::_value,Father_t::u,Father_t::u);
        }  
    }
    
    void _wait () {}

};

#endif // __BBContainer_SYMMETRIZE_H__
